from typing import Any

from pydantic import create_model
from result import Err, Ok

from ..dicts import get
from .rules import RGC, Rule, RuleGroup


def validate(
    source: dict[str, Any], validation_map: dict[str, Rule | RuleGroup]
) -> Ok[dict] | Err[RuleGroup | str]:
    """
    Performs valiudation on the `source` dict. Enforces corresponding `Rule`s and `RuleGroup`s
      at a given key
    """

    # Try applying each rule at the given key
    failed_r_rg = RuleGroup([], RGC.ALL_RULES)
    for k, r_rg in validation_map.items():
        # Run rules on source[k], return `Err` if get fails
        # Append key information when available
        if r_rg._key:
            k = f"{k}.{r_rg._key}"
        curr_level = get(source, k, default=Err(f"Failed to get key {k} from: {source}"))
        if isinstance(curr_level, Err):
            return curr_level
        # Execute rules, then append `Err` if it comes up
        res = r_rg(curr_level)
        if isinstance(res, Err):
            failed_r_rg.append(res.err_value)  # type: ignore

    if failed_r_rg:
        return Err(failed_r_rg)

    return Ok(source)


# TODO: Implement this and add to pydantic test

# def to_pydantic(validation_map: dict[str, Rule | RuleGroup]) -> dict[str, Any]:
#     """
#     Convert a validation_map to a Pydantic model.

#     OK. Try the following algo:
#     1. First-pass of validation_map:
#         - Parse-out structural schema
#         - Parse-out custom Rule/RuleGroup requirements
#     2. Use pydantic `create_model` to create custom class at runtime
#         - Validate `source` against custom class
#     3. Run additional rules from first-pass
#         - Each key information should be saved accordingly
#     """
#     return create_model(...)



import traceback
from collections.abc import Callable, Iterable, Sequence
from typing import Any

from .globs import SharedMapperState, _Global_Mapper_State_Dict
from .lib.types import DROP, KEEP, ApplyFunc, ConditionalCheck
from .lib.util import default_dsl, encode_stack_trace, flatten_list


def get(
    source: dict[str, Any] | list[Any],
    key: str,
    default: Any = None,
    apply: ApplyFunc | Iterable[ApplyFunc] | None = None,
    only_if: ConditionalCheck | None = None,
    drop_level: DROP | None = None,
    flatten: bool = False,
    strict: bool | None = None,
) -> Any:
    """
    Gets a value from the source dictionary using a `.` syntax.
    Handles None-checking (instead of raising error, returns default).

    `key` notes:
     - Use `.` to chain gets
     - Index and slice into lists, e.g. `[0]`, `[-1]`, `[:1]`, etc.
     - Iterate through a list using `[*]`
     - Get multiple items using `(firstKey,secondKey)` syntax (outputs as a tuple)
       The keys within the tuple can also be chained with `.`

    Optional param notes:
    - `default`: Return value if `key` results in a `None` (before other params apply)
    - `apply`: Use to safely chain operations on a successful get
    - `only_if`: Use to conditionally decide if the result should be kept + `apply`-ed.
    - `drop_level`: Use to specify conditional dropping if get results in None.
    - `flatten`: Use to flatten the final result (e.g. nested lists)
    - `strict`: Use to throw `ValueError` instead of returning `None` (also available at `Mapper`-level)
    """
    # Grab context from `Mapper` classes (if relevant)
    #   This exists to allow for `strict` mode to be set for all `get` calls that the `Mapper` has
    #   (without needing to specify it for every get call)
    mapper_state = _get_global_mapper_config()
    # For `strict`, prefer Mapper setting or take local setting
    strict = (mapper_state.strict if mapper_state else None) or strict

    if source:
        res = _nested_get(source, key, default)
        if strict:
            _enforce_strict(res, strict, key, source)
    else:
        res = default

    if flatten and isinstance(res, list):
        res = flatten_list(res)

    if res is not None and only_if:
        res = res if only_if(res) else None
        if strict:
            _enforce_strict(res, strict, key, source)

    if res is not None and apply:
        if not isinstance(apply, Iterable):
            apply = (apply,)
        for fn in apply:
            try:
                res = fn(res)
                if strict:
                    _enforce_strict(res, strict, key, source)
            except Exception as e:
                raise RuntimeError(f"`apply` call {fn} failed for value: {res} at key: {key}, {e}")
            if res is None:
                break

    if drop_level and res is None:
        res = drop_level

    return res


def _enforce_strict(
    res: Any, strict: bool | None, key: str, source: dict[str, Any] | list[Any]
) -> None:
    if strict and res is None:
        # Check if value is deliberately `None`, otherwise return error
        tokenized_keypath = _get_tokenized_keypath(key)
        nested_val: Any = source
        MISSING_VAL_INDICATOR = "__NOTFOUND__"
        for k in tokenized_keypath:
            if nested_val == MISSING_VAL_INDICATOR:
                break
            match k:
                case "*":
                    # TODO: handle list unwraps - here we'll just stop checking
                    nested_val = MISSING_VAL_INDICATOR
                case _:
                    nested_val = (
                        nested_val[k]
                        if (isinstance(k, int) or k in nested_val)
                        else MISSING_VAL_INDICATOR
                    )
        if nested_val is not None:
            raise ValueError(f"_Strict mode_: invalid key: {key}")


def _get_global_mapper_config() -> SharedMapperState | None:
    curr_trace = traceback.format_stack()
    # Iterate through all mappers, and check stack trace with key str
    for m_id, sms in _Global_Mapper_State_Dict.items():
        if len(curr_trace) <= sms._trace_len:
            continue
        if m_id == encode_stack_trace(curr_trace[: sms._trace_len]):
            return sms
    return None


def _nested_get(
    source: dict[str, Any] | list[Any],
    key: str | Any,
    default: Any = None,
    dsl_fn: Callable[[dict[str, Any] | list[Any], Any], Any] = default_dsl,
) -> Any:
    """
    Expects `.`-delimited string and tries to get the item in the dict.

    If using pydian defaults, the following benefits apply:
    - Tuple support

    If you use a custom `dsl_fn`, then logic is entrusted to that function (wgpcgr).
    """
    # Assume `key: str`. If not, then trust the custom `dsl_fn` to handle it
    # Handle tuple syntax (if they ask for a tuple, return a tuple)
    if isinstance(key, str) and ("(" in key and ")" in key):
        key = key.replace("(", "[").replace(")", "]")
        res = dsl_fn(source, key)
        if isinstance(res, list):
            res = tuple(res)
    else:
        res = dsl_fn(source, key)

    # DSL-independent cleanup
    if isinstance(res, list):
        res = [r if r is not None else default for r in res]
    if res is None:
        res = default

    return res


def _nested_set(
    source: dict[str, Any], tokenized_key_list: Sequence[str | int], target: Any
) -> dict[str, Any] | None:
    """
    Returns a copy of source with the replace if successful, else None.
    """
    res: Any = source
    try:
        for k in tokenized_key_list[:-1]:
            res = res[k]
        res[tokenized_key_list[-1]] = target
    except IndexError:
        return None
    return source


def _get_tokenized_keypath(key: str) -> tuple[str | int, ...]:
    """
    Returns a keypath with str and ints separated. Prefer tuples so it is hashable.

    E.g.: "a[0].b[-1].c" -> ("a", 0, "b", -1, "c")
    """
    tokenized_key = key.replace("[", ".").replace("]", "")
    keypath = tokenized_key.split(".")
    return tuple(int(k) if k.removeprefix("-").isnumeric() else k for k in keypath)


def drop_keys(source: dict[str, Any], keys_to_drop: Iterable[str]) -> dict[str, Any]:
    """
    Returns the dictionary with the requested keys set to `None`.

    If a key is a duplicate, then lookup fails so that key is skipped.

    DROP values are checked and handled here.
    """
    res = source
    seen_keys = set()
    for key in keys_to_drop:
        curr_keypath = _get_tokenized_keypath(key)
        if curr_keypath not in seen_keys:
            if v := _nested_get(res, key):
                # Check if value has a DROP object
                if isinstance(v, DROP):
                    # If "out of bounds", raise an error
                    if v.value > 0 or -1 * v.value > len(curr_keypath):
                        raise RuntimeError(f"Error: DROP level {v} at {key} is invalid")
                    curr_keypath = curr_keypath[: v.value]
                    # Handle case for dropping entire object
                    if len(curr_keypath) == 0:
                        return dict()
                if updated := _nested_set(res, curr_keypath, None):
                    res = updated
                seen_keys.add(curr_keypath)
        else:
            seen_keys.add(curr_keypath)
    return res


def impute_enum_values(source: dict[str, Any], keys_to_impute: set[str]) -> dict[str, Any]:
    """
    Returns the dictionary with the Enum values set to their corresponding `.value`
    """
    res = source
    for key in keys_to_impute:
        curr_val = _nested_get(res, key)
        if isinstance(curr_val, KEEP):
            literal_val = curr_val.value
            res = _nested_set(res, _get_tokenized_keypath(key), literal_val)  # type: ignore
    return res



from copy import deepcopy
from typing import Any

import pydian.partials as p

from .rules import RC, RGC, Rule, RuleGroup


class IsRequired(Rule):
    """
    A rule where the current field is required

    For `RuleGroup`: keep default contraint
    """

    def __init__(self, at_key: str | None = None):
        # For each rule, make it required
        super().__init__(p.not_equivalent(None), RC.REQUIRED, at_key=at_key)

    def __and__(self, other: Rule | RuleGroup | Any) -> Rule | RuleGroup:
        """
        Returns a `RuleGroup` with the `REQUIRED` constraint applied to the single rule
        """
        match other:
            case Rule():
                res = deepcopy(other)
                res._constraint = RC.REQUIRED
            case _:
                # Check callable case here (cast into a `Rule`)
                if not isinstance(other, RuleGroup) and callable(other):
                    res = Rule.init_specific(other, RC.REQUIRED)
                else:
                    res = super().__and__(other)
        return res

    def __rand__(self, other):
        return self.__and__(other)


class NotRequired(Rule):
    """
    When combined with another rule, removes the Required constraint.

    Adds a dummy rule that doesn't do anything explicitly, though might loosen RuleGroup constraints
      (e.g. `AT_LEAST_ONE` will always pass since we're adding one rule that always passes)
    """

    def __init__(self, at_key: str | None = None):
        # Initialize with dummy placeholder rule
        super().__init__(lambda _: True, at_key=at_key)

    def __and__(self, other: Rule | RuleGroup | Any) -> Rule | RuleGroup:
        """
        For a `Rule`: remove `REQUIRED`
        For a `RuleGroup`: set to `ALL_WHEN_DATA_PRESENT` -- this means it's optional, but validate if-present
        """
        match other:
            case Rule():
                res = deepcopy(other)
                if res._constraint is RC.REQUIRED:
                    res._constraint = None
            case RuleGroup():
                # Loops through and removes the `IsRequired` rule
                res = deepcopy(other)  # type: ignore
                try:
                    res.remove(IsRequired())  # type: ignore
                except ValueError:
                    pass
            case _:
                res = super().__and__(other)
        return res

    def __rand__(self, other: Rule | RuleGroup | Any):
        return self.__and__(other)


class InRange(Rule):
    def __init__(
        self, lower: int | None = None, upper: int | None = None, at_key: str | None = None
    ):
        """
        Used to check if an list is within a size range, e.g.
            [
                str
            ] & InRange(3, 5)
          is a list of 3 to 5 `str` values

        """
        match (lower, upper):
            case (int(), None):
                fn = lambda l: len(l) >= lower
            case (None, int()):
                fn = lambda l: len(l) <= upper
            case (int(), int()):
                fn = lambda l: lower <= len(l) <= upper
            case (None, None):
                raise ValueError("Need to specify lower and/or upper bound: none received!")
        super().__init__(fn, at_key=at_key)


class MaxCount(Rule):
    def __init__(
        self,
        upper: int,
        constraint: RC | None = None,
        at_key: str | None = None,
    ):
        super().__init__(p.pipe(len, p.lte(upper)), constraint, at_key)


class MinCount(Rule):
    def __init__(
        self,
        lower: int,
        constraint: RC | None = None,
        at_key: str | None = None,
    ):
        super().__init__(p.pipe(len, p.gte(lower)), constraint, at_key)


class IsType(Rule):
    def __init__(
        self,
        typ: type,
        constraint: RC | None = None,
        at_key: str | None = None,
    ):
        super().__init__(p.isinstance_of(typ), constraint, at_key)


class InSet(Rule):
    """IDEA: have this be the enum variant. E.g. one of these literals"""

    def __init__(self, s: set[Any]):
        super().__init__(p.contained_in(s))



from __future__ import (  # Used to recursively type annotate (e.g. `Rule` in `class Rule`)
    annotations,
)

import inspect
from collections.abc import Callable, Collection, Iterable
from enum import Enum
from typing import Any

from result import Err, Ok

import pydian.partials as p

from ..dicts import get


class RC(Enum):
    """
    Rule Constraint (RC): A constraint for a single `Rule` (`RuleGroup` is responsible for application)

    Optional by default (i.e. no value specified)
    """

    REQUIRED = 1
    # ONLY_IF = lambda fn: (RC.REQUIRED, fn)  # Usage: RC.ONLY_IF(some_fn)
    # ONLY_AFTER = ... # Usage: ONLY_AFTER(other_rule) ... or some way to identify...


class RGC(Enum):
    """
    RuleGroup Constraint (RGC): Constraints on top of current `RuleGroup`

    Groups of >3 explicitly required are discouraged - KISS!

    RGC takes precedent over RC (i.e. a `RGC` setting will override a `RC` setting when applicable)
    """

    # TODO: Consider conditional feature -- `ONLY_IF(...)`, `ALL_IF(...)` instead of `ALL_WHEN_DATA_PRESENT`
    #       Would need to think through if they take generic callables, Rules, etc.

    ALL_REQUIRED_RULES = -2  # NOTE: This makes rules default _optional_
    ALL_RULES = -1  # NOTE: This makes rules default _required_
    ALL_WHEN_DATA_PRESENT = 0  # NOTE: This makes rules default _required when data is present_
    AT_LEAST_ONE = 1
    AT_LEAST_TWO = 2
    AT_LEAST_THREE = 3


class Rule:
    """
    A `Rule` is a callable that either returns:
    - Ok(truthy_result)
    - Err(str + falsy_result)

    A `Rule` can have additional constraints. These will ONLY be applied if contained within a `RuleGroup`

    `Rule`s can be combined to create `RuleGroup`s using: `&`, `|`
    """

    def _raise_undefined_rule_err():  # type: ignore
        raise ValueError("Rule was not defined!")

    _fn: Callable = lambda _: Rule._raise_undefined_rule_err()
    _constraint: RC | None = None
    _key: str | None = None

    def __init__(
        self,
        fn: Callable,
        constraint: RC | None = None,
        at_key: str | None = None,
    ):
        self._fn = fn
        self._key = at_key
        if constraint:
            self._constraint = constraint

    def __call__(self, source: Any, *args) -> Ok[Any] | Err[str]:
        """
        Call specified rule and wraps as a `Result` type
        """
        # NOTE: Only apply key logic for `dict`s. Something something, design choice!
        # Also: if passed an `Ok`, unwrap it by default
        if isinstance(source, dict) and self._key:
            source = get(source, self._key)
        elif isinstance(source, Ok):
            source = source.unwrap()
        elif isinstance(source, Err):
            raise ValueError(f"Rule called with Err: {source}")
        try:
            res = self._fn(source, *args)
            if res:
                return Ok(res)
        except Exception as e:
            return Err(f"{self} failed with exception: {e}")
        return Err(f"{self} failed, got falsy value: {res}")

    def __repr__(self) -> str:
        # NOTE: can only grab source for saved files, not in repl
        #  So the function needs to be saved on a file
        #  See: https://stackoverflow.com/a/335159
        try:
            return f"<Rule> {inspect.getsource(self._fn).strip()}, ({self._fn.__closure__[0].cell_contents})"  # type: ignore
        except:
            try:
                return f"<Rule> {inspect.getsource(self._fn).strip()}"
            except:
                return f"<Rule> {self._fn}"

    def __hash__(self):
        return hash((self._fn, self._constraint, self._key))

    def __eq__(self, other: Rule | Any):
        if isinstance(other, Rule):
            # Rules are the same based on the code and the applied constraints
            # TODO: I _think_ this will work, though need to test more thoroughly
            #   e.g. this won't work for built-in callables like `str`, `bool`
            try:
                return other._fn.__code__.co_code == self._fn.__code__.co_code
            except:
                return other._fn == self._fn
        return NotImplemented

    @staticmethod
    def init_specific(v: Any, constraint: RC | None = None, at_key: str | None = None) -> Rule:
        """
        Generically returns a more specific rule when possible
        """
        # TODO: consider `InRange` (take range obj) and `InSet` (take set object)
        from .specific import IsType  # Import here to avoid circular import

        if isinstance(v, type):
            res = IsType(v, constraint, at_key)  # type: ignore
        else:
            res = Rule(v, constraint, at_key)  # type: ignore
        return res

    def __and__(self, other: Rule | RuleGroup | Any):
        return RuleGroup.combine(self, other)

    def __rand__(self, other: Rule | RuleGroup | Any):
        # Operation intended to be commutative
        return self.__and__(other)

    def __or__(self, other: Rule | RuleGroup | Any):
        return RuleGroup.combine(self, other, RGC.AT_LEAST_ONE)

    def __ror__(self, other: Rule | RuleGroup | Any):
        # Operation intended to be commutative
        return self.__or__(other)


class RuleGroup(list):
    """
    A `RuleGroup` is a callable list that can contain:
    - One or many `Rule`s
    - Other `RuleGroup`s (ending in a terminal `RuleGroup` of just `Rule`s at some point)

    The specified `RGC` (RuleGroup Constraint) determines the logic for the result outcome.
      Additionally, individual `Rule`s may have constraints which the `RuleGroup` manages
    """

    _constraint: RGC | None = None  # type: ignore
    _n_rules: int = 0
    # NOTE: _key is needed here to save nesting information from operations like `&` with a `dict`
    #   e.g. a user-specified `RuleGroup` shouldn't need to specify `_key` for each rule,
    #   rather we should infer that during parsing
    _key: str | None = None

    def __init__(
        self,
        items: Rule | RuleGroup | Callable | Collection[Rule | RuleGroup | Callable] | None = None,
        constraint: RGC = RGC.ALL_RULES,
        at_key: str | None = None,
    ):
        self._key = at_key
        self._constraint = constraint

        # Type-check and handle items
        if items:
            res = []
            # Check items
            if not isinstance(items, Iterable):
                items = [items]
            for it in items:
                if not (isinstance(it, Rule) or isinstance(it, RuleGroup)):
                    # Add a new `Rule` wrapper if applicable
                    if callable(it):
                        it = Rule.init_specific(it)
                    else:
                        raise ValueError(
                            f"All items in a `RuleGroup` must be `Rule`s or `RuleGroup`s, got: {type(it)}"
                        )
                res.append(it)
                if isinstance(it, RuleGroup):
                    self._n_rules += it._n_rules
                else:
                    self._n_rules += 1
            super().__init__(res)
        else:
            super().__init__()

    def append(self, item: Rule | RuleGroup | Callable):
        if not (isinstance(item, Rule) or isinstance(item, RuleGroup)):
            if callable(item):
                item = Rule.init_specific(item)
            else:
                raise ValueError(f"All items in a `RuleGroup` must be `Rule`s, got: {type(item)}")
        if isinstance(item, RuleGroup):
            self._n_rules += item._n_rules
        else:
            self._n_rules += 1
        super().append(item)

    def extend(self, item: RuleGroup | Iterable[Rule | RuleGroup | Callable]):
        if isinstance(item, RuleGroup):
            self._n_rules += item._n_rules
        super().extend(item)

    @staticmethod
    def combine(
        first: Rule | RuleGroup,
        other: Rule | RuleGroup | Any,
        constraint: RGC = RGC.ALL_WHEN_DATA_PRESENT,
    ) -> RuleGroup:
        """
        Combines a `RuleGroup` with another value. By default, all data is optional by default

        Here are the expected cases:
        1. & RuleGroup -> Make a new RuleGroup with both existing ones (i.e. add a nesting parent)
        2. & Rule -> Add it to current RuleGroup
        3. & type -> Add type check
        4. & dict -> Add 1) type check, 2) contents of dict
        5. & list -> Add 1) type check, 2) contents of list
        6. & Callable -> Add the callable as a Rule
        7. & some primitive -> Add an equality check for the primitive
        """
        if isinstance(other, RuleGroup):
            return RuleGroup((first, other), constraint)
        res = RuleGroup(first, constraint)
        match other:
            case Rule():
                res.append(other)
            case type():
                res.append(Rule.init_specific(other))
            case dict():
                res.append(Rule.init_specific(dict))  # Type check
                drs = _dict_to_rulegroup(other)
                res.append(drs)
            case list():
                res.append(Rule.init_specific(list))  # Type check
                lrs = _list_to_rulegroup(other)
                res.append(lrs)
            case _:
                if callable(other):
                    res.append(Rule(other))
                else:
                    res.append(Rule(p.equals(other)))
        return res

    def __call__(self, source: Any, *args) -> Ok[RuleGroup] | Err[RuleGroup]:
        """
        Calls all embedded `Rule | RuleGroup`s and returns a corresponding result -- either:
          - Ok(rules_or_rgs_passed)
          - Err(rules_or_rgs_failed)

        If the `RuleGroup` is nested: the returned `RuleGroup` will maintain the original structure.
          Otherwise: an un-nested structure is used!

        If there are nested `RuleGroup`s, then expect the inner-most failed `RuleGroup` to have
          discrete information on why the case failed (i.e. the original nested structure is kept).

          E.g. for a RuleGroup of: `[A, B, [C, D, E]]` and `B, D, E` fails,
            expect: `Err([B, [D, E]])` as the return value.

            I.e., we know that Rule `B` and RuleGroup `[C, D, E]` failed with `[D, E]` specifically
            (and that information is retained as a RuleGroup)

          E.g. for a RuleGroup of: `[A, [B, [C, D, E]]]` and `B, D, E` fails,
            expect: `Err([[B, [D, E]]])` as the return value
        """
        # Apply key unnesting logic
        # NOTE: only when source is a dict. Design choice!
        if isinstance(source, dict) and self._key:
            source = get(source, self._key)

        # Run each rule and save results
        # NOTE: This nests results in a RuleGroup by default. For recursive calls, we'll unnest this below
        rg_passed, rg_failed = RuleGroup(constraint=RGC.ALL_RULES), RuleGroup(
            constraint=RGC.ALL_RULES
        )
        for curr_item in self:
            # Run the rule(s)
            curr_res = curr_item(source, *args)
            # Handle the different cases
            match curr_item, curr_res:
                case (Rule(), Ok()):
                    rg_passed.append(curr_item)
                case (Rule(), Err()):
                    rg_failed.append(curr_item)
                case (RuleGroup(), Ok()):
                    rg_passed.append(curr_res.ok_value)
                case (RuleGroup(), Err()):
                    rg_failed.append(curr_res.err_value)
                case _:
                    raise RuntimeError(
                        f"Unexpected type or result: {type(curr_item)}, {type(curr_res)}"
                    )

        ## Check for failed required rules -- return Err early if so
        if _contains_required_rule(rg_failed):
            return Err(rg_failed)

        # Check result and return
        res: Ok[RuleGroup] | Err[RuleGroup]
        # NOTE: `_n_rules` is the total number of discrete rules (including nested).
        #      Thus, for `ALL_RULES` we check every rule, and `AT_LEAST_x` we check top-level groups
        match self._constraint:
            case RGC.ALL_RULES:
                res = Ok(rg_passed) if rg_passed._n_rules == self._n_rules else Err(rg_failed)
            case RGC.AT_LEAST_ONE | RGC.AT_LEAST_TWO | RGC.AT_LEAST_THREE:
                res = Ok(rg_passed) if len(rg_passed) >= self._constraint.value else Err(rg_failed)
            case RGC.ALL_REQUIRED_RULES:
                # Since we have above check for required rules, we know all rules have passed here
                res = Ok(rg_passed)
            case RGC.ALL_WHEN_DATA_PRESENT:
                # For each failed rule, check if data was present. If so, return `Err`
                res = Ok(rg_passed) if not _rulegroup_applies(rg_failed, source) else Err(rg_failed)
            case _:
                raise RuntimeError(f"Unsupported RuleGroup constraint: {self._constraint}")
        return res

    def __hash__(self):
        return hash(tuple(self))

    def __repr__(self) -> str:
        return f"<RuleGroup> {[r for r in self]}"

    def __and__(self, other: Rule | RuleGroup | Any):
        return RuleGroup.combine(self, other)

    def __rand__(self, other: Rule | RuleGroup | Any):
        # Expect commutative
        return self.__and__(other)

    def __or__(self, other: Rule | RuleGroup | Any):
        return RuleGroup.combine(self, other, RGC.AT_LEAST_ONE)

    def __ror__(self, other: Rule | RuleGroup | Any):
        # Expect commutative
        return self.__or__(other)


""" Helper Functions """


def _contains_required_rule(rg: RuleGroup | Rule) -> bool:
    """
    Returns `True` if `rg` contains at least 1 Rule with required constraint
    """
    if isinstance(rg, Rule):
        return rg._constraint == RC.REQUIRED
    return any(_contains_required_rule(r) for r in rg)


def _rulegroup_applies(rg: RuleGroup | Rule, source: dict[str, Any]) -> bool:
    """
    Returns `True` if `rg` applies to _any_ part of `source`
      This is mainly to handle the `ALL_WHEN_DATA_PRESENT` logic

    If there is no key-level data in `rg`, then conservatively assume there is overlap
    """
    if isinstance(rg, Rule):
        return rg._key is None or get(source, rg._key) is not None
    return any(_rulegroup_applies(r, source) for r in rg)


def _list_to_rulegroup(
    l: list[Callable | Rule | RuleGroup | dict | list], key_prefix: str | None = None
) -> RuleGroup:
    """
    Given a list, compress it into a single `RuleGroup`

    An item in the list should pass at least one of the callables in the `RuleGroup`.
      For example: `{ 'k': [ str, bool ]}` -- at key 'k', it can contain a list `str` or `bool`
      (for more specific constraints: use a nested `RuleGroup`)
    """
    # TODO: Check the key is getting applied correctly
    res = RuleGroup(constraint=RGC.AT_LEAST_ONE, at_key=key_prefix)
    for it in l:
        # TODO: does this even work / is this even needed?
        at_key = f"[*]"  # Should be applied to every item in the list
        match it:
            case Rule():
                it._key = at_key
                res.append(it)
            case RuleGroup():
                it._key = at_key
                res = it
            case dict():
                res.append(_dict_to_rulegroup(it, at_key))
            case list():
                res.append(_list_to_rulegroup(it, at_key))
            case _:
                if callable(it):
                    res.append(Rule.init_specific(it, at_key=at_key))
                else:
                    # Exact value check
                    res.append(Rule(p.equals(it), at_key=at_key))
    return res


def _dict_to_rulegroup(d: dict[str, Rule | RuleGroup], key_prefix: str | None = None) -> RuleGroup:
    """
    Given a dict, compress it into a single `RuleGroup` with key information saved

    NOTE: expect this dict to be 1-layer (i.e. not contain other dicts) -- other dicts should
      be encompassed in a `RuleGroup`
    """
    res = RuleGroup(at_key=key_prefix)
    for k, v in d.items():
        at_key = k
        match v:
            case Rule() | RuleGroup():
                v._key = at_key
                res.append(v)
            case dict():
                res.append(Rule.init_specific(dict, at_key=at_key))
                res.append(_dict_to_rulegroup(v, key_prefix=at_key))
            case list():
                res.append(Rule.init_specific(list, at_key=k))
                res.append(_list_to_rulegroup(v, key_prefix=at_key))
            case _:
                if callable(v):
                    res.append(Rule.init_specific(v, at_key=at_key))
                else:
                    # Exact value check
                    res.append(Rule(p.equals(v), at_key=k))
    return res



from collections.abc import Callable, Container, Iterable, Reversible
from functools import partial
from itertools import islice
from typing import Any, Type, TypeVar

from result import Err, Ok

import pydian
from pydian.lib.types import DROP, ApplyFunc, ConditionalCheck

"""
`pydian` Wrappers
"""


def get(
    key: str,
    default: Any = None,
    apply: ApplyFunc | Iterable[ApplyFunc] | None = None,
    only_if: ConditionalCheck | None = None,
    drop_level: DROP | None = None,
    flatten: bool | None = None,
) -> ApplyFunc:
    """
    Partial wrapper around the Pydian `get` function
    """
    kwargs = {
        "key": key,
        "default": default,
        "apply": apply,
        "only_if": only_if,
        "drop_level": drop_level,
        "flatten": flatten,
    }
    return partial(pydian.get, **kwargs)


def pipe(*funcs: ApplyFunc) -> ApplyFunc:
    """
    Custom wrapper that applies the functions in-order and returns a result
    """

    # TODO: Make this result type, and also be ergonomic for the user!
    #       Think through with pipeline module (i.e. graceful failure case)
    def run_pipe(val: Any) -> Any:
        for func in funcs:
            val = func(val)
        return val

    return partial(run_pipe)  # TODO: I don't think I need the `partial` here, test to confirm


"""
Generic Wrappers
"""


def do(func: Callable, *args: Any, **kwargs: Any) -> ApplyFunc:
    """
    Generic partial wrapper for functions.

    Starts at the second parameter when using *args (as opposed to the first).
    """
    return lambda x: func(x, *args, **kwargs)


def add(value: Any, before: bool = False) -> ApplyFunc:
    if before:
        return lambda v: value + v
    return lambda v: v + value


def subtract(value: Any, before: bool = False) -> ApplyFunc:
    if before:
        return lambda v: value - v
    return lambda v: v - value


def multiply(value: Any, before: bool = False) -> ApplyFunc:
    if before:
        return lambda v: value * v
    return lambda v: v * value


def divide(value: Any, before: bool = False) -> ApplyFunc:
    if before:
        return lambda v: value / v
    return lambda v: v / value


T = TypeVar("T", list[Any], tuple[Any])


def keep(n: int) -> ApplyFunc | Callable[[T], T]:
    return lambda it: it[:n]


def index(idx: int) -> ApplyFunc | Callable[[Reversible], Any]:
    def get_index(obj: Reversible, i: int) -> Any:
        if i >= 0:
            it = iter(obj)
        else:
            i = (i + 1) * -1
            it = reversed(obj)
        return next(islice(it, i, i + 1), None)

    return partial(get_index, i=idx)


def equals(value: Any) -> ConditionalCheck:
    # if type(value) == pl.DataFrame:
    #     return lambda df: df.equals(value)
    return lambda v: v == value


def gt(value: Any) -> ConditionalCheck:
    return lambda v: v > value


def lt(value: Any) -> ConditionalCheck:
    return lambda v: v < value


def gte(value: Any) -> ConditionalCheck:
    return lambda v: v >= value


def lte(value: Any) -> ConditionalCheck:
    return lambda v: v <= value


def equivalent(value: Any) -> ConditionalCheck:
    return lambda v: v is value


def contains(value: Any) -> ConditionalCheck:
    return lambda container: value in container


def contained_in(container: Container) -> ConditionalCheck:
    return lambda v: v in container


def not_equal(value: Any) -> ConditionalCheck:
    return lambda v: v != value


def not_equivalent(value: Any) -> ConditionalCheck:
    return lambda v: v is not value


def not_contains(value: Any) -> ConditionalCheck:
    return lambda container: value not in container


def not_contained_in(container: Container) -> ConditionalCheck:
    return lambda v: v not in container


def isinstance_of(type_: Type) -> ConditionalCheck:
    return lambda v: isinstance(v, type_)


"""
stdlib Wrappers
"""


def map_to_list(func: Callable) -> ApplyFunc | Callable[[Iterable], list[Any]]:
    """
    Partial wrapper for `map`, then casts to a list
    """
    _map_to_list: Callable = lambda fn, it: list(map(fn, it))
    return partial(_map_to_list, func)


def filter_to_list(func: Callable) -> ApplyFunc | Callable[[Iterable], list[Any]]:
    """
    Partial wrapper for `filter`, then casts to a list
    """
    _filter_to_list: Callable = lambda fn, it: list(filter(fn, it))
    return partial(_filter_to_list, func)



import base64
from collections.abc import Collection
from itertools import chain
from typing import Any, TypeVar

import jmespath

DL = TypeVar("DL", dict[str, Any], list[Any], Any)


def remove_empty_values(input: DL) -> DL:
    """
    Recursively removes "empty" objects (`None` and/or objects only containing `None` values).
    """
    if isinstance(input, list):
        return [remove_empty_values(v) for v in input if has_content(v)]
    elif isinstance(input, dict):
        return {k: remove_empty_values(v) for k, v in input.items() if has_content(v)}
    return input


def has_content(obj: Any) -> bool:
    """
    Checks if the object has "content" (a non-`None` value), and/or contains at least one item with "content".
    """
    res = obj is not None
    if res and isinstance(obj, Collection):
        res = len(obj) > 0
        # If has items, recursively check if those items have content.
        #   A case has content if at least one inner item has content.
        if isinstance(obj, list):
            res = any(has_content(item) for item in obj)
        elif isinstance(obj, dict):
            res = any(has_content(item) for item in obj.values())
    return res


def get_keys_containing_class(source: dict[str, Any], cls: type, key_prefix: str = "") -> set[str]:
    """
    Recursively finds all keys where a DROP object is found.
    """
    res = set()
    for k, v in source.items():
        curr_key = f"{key_prefix}.{k}" if key_prefix != "" else k
        match v:
            case cls():  # type: ignore
                res.add(curr_key)
            case dict():
                res |= get_keys_containing_class(v, cls, curr_key)
            case list():
                for i, item in enumerate(v):
                    indexed_keypath = f"{curr_key}[{i}]"
                    if isinstance(item, cls):
                        res.add(indexed_keypath)
                    elif isinstance(item, dict):
                        res |= get_keys_containing_class(item, cls, indexed_keypath)
    return res


def flatten_list(res: list[list[Any]]) -> list[Any]:
    """
    Flattens a list-of-list
    E.g. Given:    [[1, 2, 3], [4, 5, 6], None, [7, 8, 9]]
         Returns:  [1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    if res_without_nones := [l for l in res if (l is not None) and (isinstance(l, list))]:
        res = list(chain.from_iterable(res_without_nones))
        # Handle nested case
        res = flatten_list(res)
    return res


def default_dsl(source: dict[str, Any] | list[Any], key: str):
    """
    Specifies a DSL (domain-specific language) to use when running `get`

    Here, we redefine the `jmespath.search` to be consistent with argument ordering in the repo
    """
    return jmespath.search(key, source)


def encode_stack_trace(stack_trace: list[str]) -> str:
    # Encode the stack trace (into bytes), then save the byte representation as a str (second `decode`)
    return base64.b64encode(bytes(("".join(stack_trace)), "utf-8")).decode("utf-8")
